unit GridList;

interface

{ TGridList : something between a StringGrid and a ListBox...

Copyright (C) 1998 TothPaul@Mygale.org
 http://www.mygale.org/~tothpaul

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, ExtCtrls;

type
  pIntegerList=^TIntegerList;
  TIntegerList=array[0..MaxListSize] of Integer; // handle Headers width

  pLine=^TLine;       // List Items
  TLine=record
   Style:TFontStyles;
   Lines:TStringList;
  end;

  TGridList = class(TCustomControl)
  private
    fHeaders:TStringList;          // Headers
    fHeadSize:integer;             // Headers Height
    fHeaderFont:TFont;
    fHeaderWidth:pIntegerList;
    fHeaderCount:word;
    fHitTest:TPoint;               // use to manage OnClick & Header Width
    fCanResize:boolean;
    fResizeSection:integer;
    fMouseOffset:integer;
    fMaxHit:integer;
    fTopPos:integer;               // current vertical scrollpos
    fLeftPos:integer;              // current horizontal scrollpos
    fRange:TPoint;                 // ScrollBar ranges
    fItems:TList;
    fItemSize:Integer;             // Item Height
    fSelected:Integer;             // current item
    FBorderStyle: TBorderStyle;
    procedure SetHeaders(Value:TStringList);
    procedure SetHeaderFont(Value:TFont);
    procedure RepaintOnChange(Sender:TObject);
    procedure OnHeadersChange(Sender:TObject);
    procedure SetHeaderWidth(Index,Value:integer);
    function  GetHeaderWidth(Index:integer):integer;
    procedure SetItem(Index:integer; Value:TStrings);
    function GetItem(Index:integer):TStrings;
    function GetCurrent:TStrings;
    procedure SetStyle(Index:Integer; Value:TFontStyles);
    function GetStyle(Index:Integer):TFontStyles;
    function GetItemCount:integer;
    procedure SetSelected(Value:Integer);
    procedure SetBorderStyle(Value: TBorderStyle);
   // resizing (from THeader)
    procedure WMSetCursor(var Msg: TWMSetCursor); message WM_SETCURSOR;
    procedure WMNCHitTest(var Msg: TWMNCHitTest); message WM_NCHITTEST;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
   // scrollbars (from my THTMemo)
    procedure WMSize(Var Msg:TWMSize); message wm_size;
    procedure WMVScroll(Var Msg:TWMVScroll); message wm_vscroll;
    procedure WMHScroll(Var Msg:TWMHScroll); message wm_hscroll;
    procedure SetTopPos(Value:Integer);
    procedure SetLeftPos(Value:Integer);
    procedure SetCell(x,y:integer; Value:string);
    function GetCell(x,y:integer):string;
   // load/save
    procedure ReadData(Reader: TReader);
    procedure WriteData(Writer: TWriter);
   // lines
    function NewLine(Style:TFontStyles; s:TStrings):pLine;
    procedure FreeLine(Line:PLine);
  protected
    procedure Paint; override;
    procedure DefineProperties(Filer: TFiler); override;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure ScrollRange;
    procedure InvalidateItems;
    procedure InvalidateItem(Index:integer);
    procedure SelectItem(Index:Integer);
    function HeadRect:TRect;
    function ItemsRect:TRect;
   // limit Click on available items
    procedure Click; override;
    procedure DblClick; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    Procedure Clear;
    Property HeaderWidth[Index:integer]:integer read GetHeaderWidth write SetHeaderWidth;
    Property Items[Index:Integer]:TStrings read GetItem write SetItem;
    Property Style[Index:Integer]:TFontStyles read GetStyle write SetStyle;
    Property ItemCount:integer read GetItemCount;
    Procedure AddItem(Strings:Array of String);
    Procedure Delete(Index:integer);
    procedure AutoSize(index:integer);
    Property TopPos:Integer read FTopPos write SetTopPos;
    Property LeftPos:Integer read FLeftPos write SetLeftPos;
    Property Selected:Integer read fSelected write SetSelected;
    Property Cells[x,y:Integer]:string read GetCell write SetCell;
    Property Current:TStrings read GetCurrent;
  published
    Property Headers:TStringList read fHeaders write SetHeaders;
    Property HeaderFont:TFont read fHeaderFont write SetHeaderFont;
    property BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle default bsSingle;
    Property Align;
    Property Color;
    Property ParentColor;
    Property Font;
    Property ParentFont;
    Property Hint;
    Property ShowHint;
    Property ParentShowHint;
    Property PopupMenu;
    Property Ctl3D;
    Property ParentCtl3D;
    Property OnClick;
    Property OnDblClick;
    Property TabStop;
    Property TabOrder;
    Property OnKeyDown;
    Property OnKeyUp;
    Property OnKeyPress;
    Property Enabled;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('MySoft', [TGridList]);
end;

// GridList

constructor TGridList.Create(AOwner: TComponent);
 begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csDesignInteractive, csOpaque, csFramed];
  FBorderStyle := bsSingle;
  fItemSize:=11;
  fSelected:=-1;
  fHeaderWidth:=nil;
  fHeaderCount:=0;
  fHeaders:=TStringList.Create;
  fItems:=TList.Create;
  fHeaders.OnChange:=OnHeadersChange;
  fHeaderFont:=TFont.Create;
  fHeaderFont.OnChange:=RepaintOnChange;
  fMaxHit:=0;
  ParentColor:=false;
  Color:=clWindow;
  TabStop:=True;
 end;

procedure TGridList.CreateParams(var Params: TCreateParams);
 Const
  BorderStyles: array[TBorderStyle] of Longint = (0, WS_BORDER);
 begin
  inherited CreateParams(Params);
  Params.Style:=Params.Style or WS_HSCROLL or WS_VSCROLL or BorderStyles[FBorderStyle];
 end;

destructor TGridList.Destroy;
 begin
  fHeaderFont.Free;
  fHeaders.Free;
  inherited Destroy;
 end;

procedure TGridList.DefineProperties(Filer: TFiler);
begin
  if Filer is TReader then inherited DefineProperties(Filer);
  Filer.DefineProperty('HeaderWidth', ReadData, WriteData, fHeaderCount > 0);
end;

procedure TGridList.WriteData(Writer:TWriter);
 var
  i:integer;
 begin
  Writer.WriteListBegin;
  Writer.WriteInteger(fHeaderCount);
  for i:=0 to fHeaderCount-1 do Writer.WriteInteger(fHeaderWidth[i]);
  Writer.WriteListEnd;
 end;

procedure TGridList.ReadData(Reader:TReader);
 var
  i:integer;
 begin
  Reader.ReadListBegin;
  fHeaderCount:=Reader.ReadInteger;
  GetMem(fHeaderWidth,fHeaderCount*SizeOf(Integer));
  for i:=0 to fHeaderCount-1 do fHeaderWidth[i]:=Reader.ReadInteger;
  Reader.ReadListEnd;
 end;

procedure TGridList.RepaintOnChange(Sender:TObject);
 begin
  if not (csLoading in componentstate) then Paint;
 end;

procedure TGridList.OnHeadersChange(Sender:TObject);
 var
  i:integer;
  h:pIntegerList;
 begin
  if fHeaders.Count<>fHeaderCount then begin  // need to update Widths
   if fHeaders.Count=0 then  // how many header
    h:=nil
   else begin
    getmem(h,fHeaders.Count*SizeOf(Integer));   // allocate new array
    if fHeaderCount>0 then begin   // copy old values
     if fHeaderCount> fHeaders.Count then i:=fHeaders.Count else i:=fHeaderCount;
     move(fHeaderWidth^,h^,i*SizeOf(Integer));
    end;
    for i:=fHeaderCount to fHeaders.Count-1 do // set a defaut value for new records
     h^[i]:=Canvas.TextWidth(fHeaders[i])+5;
   end;
  // free old array
   if fHeaderCount>0 then freemem(fHeaderWidth,fHeaderCount*SizeOf(Integer));
  // refer new array
   fHeaderWidth:=h;
   fHeaderCount:=fHeaders.Count;
  end;
  if not (csLoading in componentstate) then Paint;
 end;

function TGridList.NewLine(Style:TFontStyles; s:TStrings):pLine;
 begin
  New(Result);
  Result.Style:=Style;
  Result.Lines:=TStringList.Create;
  Result.Lines.OnChange:=RepaintOnChange;
  Result.Lines.Assign(s);
 end;

procedure TGridList.FreeLine(Line:PLine);
 begin
  TStringList(Line.Lines).Free;
  Dispose(Line);
 end;

procedure TGridList.SetBorderStyle(Value: TBorderStyle);
begin
  if Value <> FBorderStyle then begin
   FBorderStyle := Value;
   RecreateWnd;
  end;
end;

procedure TGridList.Clear;
 var
  i:integer;
 begin
  for i:=0 to fItems.Count-1 do FreeLine(fItems[i]);
  fItems.clear;
  fSelected:=-1;
  Invalidate;
 end;

Procedure TGridList.AddItem(Strings:Array of String);
 var
  i:integer;
  s:TStrings;
  l:pLine;
 begin
  s:=TStringList.Create;
  for i:=0 to High(Strings) do s.add(Strings[i]);
  Items[fItems.Count]:=s;
  s.free;
 end;

Procedure TGridList.Delete(Index:Integer);
 begin
  if (Index<0)or(Index>=FItems.count) then exit;
  FreeLine(fItems[Index]);
  fItems.Delete(Index);
  Invalidate;
 end;

Procedure TGridList.AutoSize(Index:integer);
 var
  NewWidth:integer;
  i,w:integer;
 begin
  if Index>fHeadercount then exit;
  NewWidth:=Canvas.TextWidth(Headers[Index]);
  for i:=0 to fItems.Count-1 do begin
   if Items[i].Count>Index then begin
    w:=Canvas.TextWidth(Items[i][Index]);
    if w>NewWidth then NewWidth:=w;
   end;
  end;
  Inc(NewWidth,8);
  if NewWidth<>HeaderWidth[index] then HeaderWidth[index]:=NewWidth;
 end;

Procedure TGridList.SetSelected(Value:Integer);
 begin
  InvalidateItem(fSelected);
  fSelected:=Value;
  InvalidateItem(fSelected);
 end;

procedure TGridList.SetHeaders(Value:TStringList);
 begin
  fHeaders.Assign(Value);
 end;

procedure TGridList.SetHeaderFont(Value:TFont);
 begin
  fHeaderFont.Assign(Value);
  if not (csLoading in componentstate) then invalidate;
 end;

procedure TGridList.SetHeaderWidth(Index,Value:integer);
 begin
  if (Index>=0)and(Index<fHeaderCount) then begin
   fHeaderWidth^[Index]:=Value;
   Invalidate;
   ScrollRange;
  end;
 end;

function TGridList.GetHeaderWidth(index:integer):integer;
 begin
  if (Index>=0)and(Index<fHeaderCount) then Result:=fHeaderWidth^[Index] else Result:=0;
 end;

procedure TGridList.SetItem(Index:Integer; Value:TStrings);
 begin
  if Index>=fItems.Count then begin
   Index:=fItems.Count;
   fItems.add(NewLine(Font.Style,Value));
   ScrollRange;
  end else
   TLine(fItems[Index]^).Lines.Assign(Value);
  if fRange.y>0 then TopPos:=fRange.y else InvalidateItem(Index);
 end;

Function TGridList.GetItem(Index:Integer):TStrings;
 begin
  while Index>=fItems.Count do AddItem(['']);
  if (Index>=0) then
   Result:=TLine(fItems[Index]^).Lines
  else
   Result:=nil;
 end;

Function TGridList.GetCurrent:TStrings;
 begin
  if fItems.Count=0 then AddItem(['']);
  if FSelected<0 then FSelected:=0;
  Result:=TLine(fItems[FSelected]^).Lines;
 end;

Procedure TGridList.SetStyle(Index:Integer; Value:TFontStyles);
 begin
  if (Index>=0)and(index<fItems.Count) then begin
   TLine(fItems[Index]^).Style:=Value;
   InvalidateItems;
  end;
 end;

Function TGridList.GetStyle(Index:Integer):TFontStyles;
 begin
  if (Index>=0)and(index<fItems.Count) then
   Result:=TLine(fItems[Index]^).Style
  else
   Result:=[];
 end;

Function TGridList.GetItemCount:integer;
 begin
  Result:=fItems.Count;
 end;

procedure TGridList.SetCell(x,y:integer; Value:string);
 var
  s:TStrings;
 begin
  s:=Items[y];
  while x>=s.count do s.add('');
  Items[y][x]:=Value;
 end;

function TGridList.GetCell(x,y:integer):string;
 begin
  if y<fItems.Count then
   Result:=Items[y][x]
  else
   Result:='';
 end;

procedure TGridList.Paint;
 var
  s:string;
  r:TRect;
  i,j:integer;
  strings:TStrings;
 begin
  if (csLoading in componentState) then exit;
  with Canvas do begin
   // headers
    if fHeaders.Count=0 then
     fHeadsize:=0
    else begin
     Font.Assign(fHeaderFont);
     fHeadSize:=TextHeight('A')+5;
     r:=HeadRect;//ClientRect;
     Dec(r.Left,fLeftPos);
     Brush.Color:=clSilver;
     for i:=0 to fHeaders.count-1 do begin
      s:=fHeaders[i];
      r.Right:=r.Left+HeaderWidth[i];
      if r.Right>ClientWidth then r.Right:=ClientWidth;
      Frame3D(Canvas,r,clWhite,clBlack,1);
      FillRect(r);
      TextRect(r,r.left,r.top,s);
      r.Left:=r.Right+1;
      if r.Left>ClientWidth then break;
      dec(r.Top);
      inc(r.Bottom);
     end;
     r.Right:=ClientWidth+1;
     Frame3D(Canvas,r,clWhite,clBlack,1);
     FillRect(r);
    end;
    fMaxHit:=fHeadSize;
   // Background
    Brush.Color:=Color;
    Brush.Style:=bsSolid;
    FillRect(ItemsRect);
   // Lines
   if fItems.Count>0 then begin
    Font.Assign(Self.Font);
    fItemSize:=TextHeight('T');
    r:=ItemsRect;
    Brush.Style:=bsClear;
    for i:=fTopPos to fItems.Count-1 do begin
     if r.Top>ClientHeight then exit; // bottom of component
     r.Bottom:=r.Top+fItemSize;
     fMaxHit:=r.Bottom;
     if i=Selected then begin
      Brush.Style:=bsSolid;
      Brush.Color:=clSilver;
      FillRect(Rect(0,r.Top,ClientWidth,r.Bottom));
      Brush.Style:=bsClear;
     end;
     r.Left:=-fLeftPos;
     With TLine(fItems[i]^) do begin
      Font.Style:=Style;
      Strings:=Lines;
     end;
     for j:=0 to Strings.Count-1 do begin
      if (j=fHeaders.Count) then break; // hidden fonction :)
      r.Right:=r.Left+fHeaderWidth[j]-8;
      if r.Right>=ClientWidth then r.Right:=ClientWidth-1;
      TextRect(r,r.left,r.top,Strings[j]);
      r.Left:=r.Right+1+8;
      if r.Left>ClientWidth then break;
     end;
     r.Top:=r.Bottom; // next line
    end;  { for fTopPos ... }
   end; { fItemsCount>0 }
  end; { with canvas }
 end;

Procedure TGridList.ScrollRange;
 var
  i:integer;
 begin
 // compute page size
  fRange.x:=0;
//  fPageWidth:=0;
  for i:=0 to fHeaderCount-1 do inc(fRange.x{PageWidth},fHeaderWidth[i]);
 // setup scrollbars
  if FRange.x{PageWidth}<ClientWidth then begin
   FRange.x:=0;
   LeftPos:=0;
  end else
   Dec(FRange.x,{:=FPageWidth-}ClientWidth);
  SetScrollRange(Handle,sb_Horz,0,FRange.x,False);

//  fPageHeight:=
  FRange.y:=(fItems.Count)-(ClientHeight-fHeadSize) div fItemSize;
//  FPageHeight;
  SetScrollRange(Handle,sb_Vert,0,FRange.y,False);
 end;

procedure TGridList.InvalidateItems;
 var
  r:TRect;
 begin
  r:=ItemsRect;
  InvalidateRect(Handle,@r,false);
 end;

procedure TGridList.InvalidateItem(Index:integer);
 var
  r:TRect;
 begin
  r.Top:=fHeadSize+(Index-fTopPos)*fItemSize;
  r.Left:=0;
  r.Bottom:=r.Top+fItemSize+1;
  r.Right:=ClientWidth;
  InvalidateRect(Handle,@r,false);
 end;

procedure TGridList.SelectItem(Index:Integer);
 begin
  if (Index>=0)and(Index<fItems.Count) then Selected:=Index;
 end;

function TGridList.HeadRect:TRect;
 begin
  Result.Top:=0;
  Result.Left:=0;
  Result.Right:=ClientWidth;
  Result.Bottom:=fHeadSize;
 end;

function TGridList.ItemsRect:TRect;
 begin
  Result.Top:=fHeadSize+1;
  Result.Left:=0;
  Result.Right:=ClientWidth;
  Result.Bottom:=ClientHeight;
 end;

procedure TGridList.Click;
 begin
  if (fHitTest.y>fHeadSize)and(fHitTest.y<fMaxHit) then inherited click;
 end;

procedure TGridList.DblClick;
 begin
  if fHitTest.y<fHeadSize then begin
   if fCanResize then begin
    fCanResize:=False;
    AutoSize(fResizeSection);
   end;
  end else
   if (fHitTest.y<fMaxHit) then inherited DblClick;
 end;

// resizing (from THeader) ----------------------------------------

procedure TGridList.WMNCHitTest(var Msg: TWMNCHitTest);
begin
  inherited;
  if Msg.Result=HTCLIENT then begin
   fHitTest := SmallPointToPoint(Msg.Pos); { locate mouse }
   fHitTest:=ScreenToClient(fHitTest); { localize mouse }
//   if fHitTest.y>fMaxHit then Msg.Result:=HTNOWHERE;
  end;
end;

procedure TGridList.WMSetCursor(var Msg: TWMSetCursor);
var
  Cur: HCURSOR;
  I: Integer;
  X: Integer;
begin
  Cur := 0;
  FResizeSection := 0;
  x := -fLeftPos;
  if fHitTest.y<fHeadSize then begin
   with Msg do
     if HitTest = HTCLIENT then { inside client }
       for I := 0 to fHeaderCount - 1 do begin
         Inc(x, fHeaderWidth[i]); { count headers width }
         fMouseOffset := fHitTest.X - x;
         if Abs(FMouseOffset) < 4 then begin { "between" two header }
           Cur := LoadCursor(0, IDC_SIZEWE);
           fResizeSection := I;
           Break;
         end;
       end;
  end;
  fCanResize := (Cur <> 0);
  if FCanResize then SetCursor(Cur) else inherited;
end;

procedure TGridList.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  if fCanResize then begin
   if ((csDesigning in ComponentState) and (Button = mbRight)) or (Button = mbLeft) then
    SetCapture(Handle);
  end else
   if (y>fHeadSize)and (not (csDesigning in ComponentState)) then begin
    y:=(y-fHeadSize)div fItemSize+fTopPos;
    SelectItem(y)
   end;
end;

procedure TGridList.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  I: Integer;
  AbsPos: Integer;
  MinPos: Integer;
  MaxPos: Integer;
  r:TRect;
begin
  inherited MouseMove(Shift, X, Y);
  if (GetCapture = Handle) and fCanResize then  begin
    { absolute position of this item }
    AbsPos := -fLeftPos;
    for I := 0 to fResizeSection do Inc(AbsPos, fHeaderWidth[i]);

    if FResizeSection > 0 then
     MinPos := AbsPos - fHeaderWidth[fResizeSection]
    else
     MinPos := 2;

    MaxPos := ClientWidth - 2;
    if X < MinPos then X := MinPos;
    if X > MaxPos then X := MaxPos;

    x:=(x-fMouseOffset)-AbsPos; // (AbsPos - X ) - FMouseOffset;
    Inc(fHeaderWidth[fResizeSection],x);
    ScrollRange;
// Invalidate;
    r:=ClientRect;
    if x>0 then begin // Extend
     r.Left:=AbsPos-8;
     ScrollWindowEx(Handle,x,0,nil,@r,0,@r,0);
     Inc(r.Right,8);
     InvalidateRect(Handle,@R,False)
    end else begin   // Reduice (x<0)
     r.Left:=AbsPos+x-6;
     dec(r.Right);
     ScrollWindowEx(Handle,x,0,nil,@r,0,@r,0);
     inc(r.Right);
     InvalidateRect(Handle,@R,False)
    end;
  end;
end;

procedure TGridList.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  if FCanResize then begin
    ReleaseCapture;
    FCanResize := False;
  end;
//  inherited MouseUp(Button, Shift, X, Y);
end;

// scrolling

procedure TGridList.WMSize(Var Msg:TWMSize);
 begin
  ScrollRange;
 end;

procedure TGridList.WMVScroll(Var Msg:TWMVScroll);
 begin
   case msg.ScrollCode of
    sb_LineUp        : TopPos:=TopPos-1;
    sb_LineDown      : TopPos:=TopPos+1;
    sb_PageUp        : TopPos:=TopPos-(ClientHeight div 2);
    sb_PageDown      : TopPos:=TopPos+(ClientHeight div 2);
    sb_Top           : TopPos:=0;
    sb_Bottom        : TopPos:=fRange.y;
    sb_ThumbTrack,
    sb_ThumbPosition : TopPos:=msg.Pos;
   end;
 end;

procedure TGridList.WMHScroll(Var Msg:TWMHScroll);
 begin
   case msg.ScrollCode of
    sb_LineUp        : LeftPos:=LeftPos-(Clientwidth div 20);
    sb_LineDown      : LeftPos:=LeftPos+(Clientwidth div 20);
    sb_PageUp        : LeftPos:=LeftPos-(ClientWidth div 2);
    sb_PageDown      : LeftPos:=LeftPos+(ClientWidth div 2);
    sb_Top           : LeftPos:=0;
    sb_Bottom        : LeftPos:=fRange.x;
    sb_ThumbTrack,
    sb_ThumbPosition : LeftPos:=Msg.Pos;
   end;
 end;

Procedure TGridList.SetTopPos(Value:integer);
 var
  R:TRect;
 begin
  if Value<0 then Value:=0 else if Value>FRange.y then Value:=FRange.y;
  if (FTopPos<>Value) then begin
   { ScrollWindowEx(Handle,dx,dy,What,Clip,UpdateHandle,UpdateRect,Flags) }
   R:=ItemsRect;
   ScrollWindowEx(Handle,0,(FTopPos-Value)*fItemSize,nil,@R,0,@R,0);
   dec(r.Top,fItemSize); // cause a new item isn't drawn & the Scroll scroll an emty line
   InvalidateRect(Handle,@R,False);
   FTopPos:=Value;
   SetScrollPos(Handle,sb_Vert,FTopPos,True);
  end;
 end;

Procedure TGridList.SetLeftPos(Value:Integer);
 var
  R:TRect;
 begin
  if Value<0 then Value:=0 else if Value>FRange.x then Value:=FRange.x;
  if (FLeftPos<>Value) then begin
 { ScrollWindowEx(Handle,dx,dy,What,Clip,UpdateHandle,UpdateRect,Flags) }
   r:=ClientRect;
   dec(r.Right);
   ScrollWindowEx(Handle,FLeftPos-Value,0,nil,@r,0,@R,0);
   inc(r.Right);
   InvalidateRect(Handle,@R,False);
   FLeftPos:=Value;
   SetScrollPos(Handle,sb_Horz,FLeftPos,True);
  end;
 end;

end.
