unit ThrdSock;

{
  Native Delphi 2 components to access IP sockets with threads
  Copyright (C) 1998  Paul Toth <tothpaul@mygale.org>
  http://www.mygale.org/~tothpaul

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

}


interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls;

Const // Thread messages
 wm_threaddone   = wm_user + 0;
 wm_setstatus    = wm_user + 1;
 wm_setlines     = wm_user + 2;
 wm_connect      = wm_user + 3;
 wm_connect_fail = wm_user + 4;
 wm_disconnect   = wm_user + 5;
 wm_callback     = wm_user + 6;

type
  TCallBack = procedure of object; // Thread call back function

// NewsGroup Document
  TNWS=Class(TComponent)
   Private
    fSource:TStrings;
    Procedure SetSource(Value:TStrings);
   public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
   Published
    Property Source:TStrings read fSource write SetSource;
  end;

// Main component (general socket access)
  TThrdSock = class(TComponent)
  private
    fHandle:THandle;
    fActive:boolean;
    fServer:string;
    fPort  :word;
    fStatus:string;
    fLines:TStrings;
    fEnabled:boolean;
    CallBack:TCallBack;
    HThread:THandle;
    ThreadId:THandle;
    HSocket:THandle;
    EOnConnect:TNotifyEvent;
    EOnConnectFail:TNotifyEvent;
    EOnDisconnect:TNotifyEvent;
    EOnBusy:TNotifyEvent;
    EOnReady:TNotifyEvent;
    EStatusChange:TNotifyEvent;
    procedure SetActive(Value:boolean);
    procedure SetServer(Value:string);
    procedure SetPort(Value:word);
    procedure SetLines(Value:TStrings);
    procedure SetStatus(Value:string);
    procedure WndProc(Var Message:TMessage);
    procedure wmThreadDone(Var Msg:TMessage); message wm_threaddone;
    procedure wmSetStatus(Var Msg:TMessage); message wm_setstatus;
    procedure wmSetLines(Var Msg:TMessage); message wm_setlines;
    procedure wmConnect(Var Msg:TMessage); message wm_connect;
    procedure wmConnectFail(Var Msg:TMessage); message wm_connect_fail;
    procedure wmDisconnect(Var Msg:TMessage); message wm_disconnect;
    procedure wmCallBack(Var Msg:TMessage); message wm_callback;
  protected
    procedure SetEnabled(Value:boolean);
    procedure Echo(Cmd:String);
    procedure RequestStatus(ACallBack:TCallBack);
    procedure RequestMessage(ACallBack:TCallBack);
  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    procedure Connect; virtual;
    procedure Disconnect;
    procedure CancelRequest;
    function  ServerAddr:longint;
    Property  Active:boolean read fActive write SetActive;
    Property Status:string read fStatus write SetStatus;
    Property Lines:TStrings read fLines write SetLines;
  published
    Property Server:string read fServer write SetServer;
    Property Port:word read fPort write SetPort;
    Property OnConnect:TNotifyEvent read EOnConnect write EOnConnect;
    Property OnConnectFail:TNotifyEvent read EOnConnectFail write EOnConnectFail;
    Property OnDisconnect:TNotifyEvent read EOnDisconnect write EOnDisconnect;
    Property OnBusy:TNotifyEvent read EOnBusy write EOnBusy;
    Property OnReady:TNotifyEvent read EOnReady write EOnReady;
    Property StatusChange:TNotifyEvent read EStatusChange write EStatusChange;
  end;

// POP3 socket component
 TThrdPOP = class(TThrdSock)
  private
   fUser:string;
   fPassword:string;
   fMailCount:integer;
   EOnMailCount:TNotifyEvent;
   EOnMail:TNotifyEvent;
   procedure wmConnect(Var Msg:TMessage); message wm_connect;
  protected
   procedure UserSelected;
   procedure LoggedIn;
   procedure MailStatus;
   procedure GotMail;
   procedure MailRemoved;
  public
   constructor Create(AOwner:TComponent); override;
   Function GetMail(Index:integer):boolean;
   Function DelMail(Index:integer):boolean;
   Property MailCount:integer read fMailCount;
  published
   Property User:string read fUser write fUser;
   Property Password:string read fPassword write fPassword;
   Property OnMailCount:TNotifyEvent read EOnMailCount write EOnMailCount;
   Property OnMail:TNotifyEvent read EOnMail write EOnMail;
 end;

 TThrdNNTP=Class(TThrdSock)
  private
   fGroups:TStrings;
   fGroup:string;
   fCount:integer;
   fFirst:integer;
   fLast :integer;
   fSubject  :string;
   fFrom     :string;
   fDate     :string;
   fLineCount:string;
   fMessageID:string;
   fReferences:TStrings;
   EGotList:TNotifyEvent;
   EOnGroupStatus:TNotifyEvent;
   EOnHeader:TNotifyEvent;
   EOnBody:TNotifyEvent;
   EOnArticle:TNotifyEvent;
   procedure SetGroups(Value:TStrings);
   procedure SetGroup(Value:string);
  protected
   procedure GroupListed;
   procedure GroupStatus;
   procedure GotHeader;
   procedure GotBody;
   procedure GotArticle;
  public
   constructor Create(AOwner:TComponent); override;
   destructor Destroy; override;
   procedure ListGroup;
   Property Groups:TStrings read fGroups write SetGroups;
   Property Group:string read fGroup write SetGroup;
   Function GetHeader(Index:integer):boolean;
   Property Count:integer read fCount;
   Property First:integer read fFirst;
   Property Last :integer read fLast;
   Property Subject:string read fSubject;
   Property From:string read fFrom;
   Property Date:string read fDate;
   Property LineCount:string read fLineCount;
   Property MessageID:string read fMessageID;
   Property References:TStrings read fReferences;
   Function GetBody:boolean;
   Function GetArticle(Id:string):boolean;
  published
   Property GotList:TNotifyEvent read EGotList write EGotList;
   Property OnGroupStatus:TNotifyEvent read EOnGroupStatus write EOnGroupStatus;
   Property OnHeader:TNotifyEvent read EOnHeader write EOnHeader;
   Property OnBody:TNotifyEvent read EOnBody write EOnBody;
   Property OnArticle:TNotifyEvent read EOnArticle write EOnArticle;
 end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('MySoft', [TThrdSock, TThrdPOP, TThrdNNTP,TNWS]);
end;

// Document
Constructor TNWS.Create(AOwner:TComponent);
 begin
  inherited Create(AOwner);
  fSource:=TStringList.Create;
 end;

Destructor TNWS.Destroy;
 begin
  fSource.free;
  inherited Destroy;
 end;

Procedure TNWS.SetSource(Value:TStrings);
 begin
  fSource.assign(Value);
 end;

//-----------------------------------------------
// Winsock APIs
//-----------------------------------------------
Const
 winsock='wsock32.dll';
 fmInternet = 2;
 scStream   = 1;
 fIoNbRead = $4004667F;

Type
  TWSAData = packed record
    wVersion: Word;
    wHighVersion: Word;
    szDescription: array[0..256] of Char;
    szSystemStatus: array[0..128] of Char;
    iMaxSockets: Word;
    iMaxUdpDg: Word;
    lpVendorInfo: PChar;
  end;

  TSockAddr=packed record
   Family:word;
   Port  :word;
   Addr  :LongInt;
   Zeros :array[0..7] of byte;
  end;

Function WSAStartup(Version:word; Var Data:TwsaData):integer; stdcall; far; external winsock;
Function WSACleanup:integer; stdcall; far; external winsock;
Function socket(Family,Kind,Protocol:integer):integer; stdcall; far; external winsock;
Function WSAIsBlocking:boolean; stdcall; far; external winsock;
Function WSACancelBlockingCall:integer; stdcall; far; external winsock;
Function shutdown(Socket,How:Integer):integer; stdcall; far; external winsock;
Function closesocket(socket:Integer):integer; stdcall; far; external winsock;
Function gethostbyname(HostName:PChar):pointer; stdcall; far; external winsock;
Function connect(socket:Integer; Var SockAddr:TSockAddr; AddrLen:integer):integer; stdcall; far; external winsock;
Function recv(socket:integer; data:pchar; datalen,flags:integer):integer; stdcall; far; external winsock;
Function send(socket:integer; data:pchar; datalen,flags:integer):integer; stdcall; far; external winsock;
Function ioctlsocket(socket:integer; cmd:integer; var reply:dword):integer; stdcall; far; external winsock;
Function WSAGetLastError:integer; stdcall; far; external winsock;

//-----------------------------------------------
// Thread Procs
//-----------------------------------------------

// Retreive a single line of text
procedure GetSocketText(Sender:TThrdSock; Notify:integer);
 var
  HSocket:THandle;
  Handle:THandle;
  Buffer:array[0..1024] of char;
  Size:integer;
  cr:pchar;
  Status:string;
 begin
  Handle :=Sender.fHandle;
  HSocket:=Sender.HSocket;
  Status :='';
  repeat
   repeat Size:=recv(HSocket,Buffer,SizeOf(Buffer)-1,0) until size<>0;
   if size=-1 then begin
    PostMessage(Handle,wm_threaddone,0,0);
    PostMessage(Handle,wm_disconnect,0,wsaGetLastError);
    exit;
   end;
   Buffer[Size]:=#0;
   cr:=StrPos(Buffer,#13);
   if cr<>nil then cr^:=#0;
   Status:=Status+Buffer;
  until cr<>nil;
  SendMessage(Handle,wm_SetStatus,0,longint(Status));
  PostMessage(Handle,wm_threaddone,0,0);
  PostMessage(Handle,Notify,0,0);
 end;

function RequestStatusThread(Sender:TThrdSock):integer; stdcall;
 begin
  GetSocketText(Sender,wm_callback);
 end;

// open socket, notify Sender and exit on error
function ConnectThread(Sender:TThrdSock):integer; stdcall;
 var
  SockAddr:TSockAddr;
 begin
  Sender.HSocket:=Socket(fmInternet,scStream,0);
  if Sender.HSocket=-1 then begin
   PostMessage(Sender.fHandle,wm_threaddone,0,0);
   PostMessage(Sender.fHandle,wm_connect_fail,wsaGetLastError,0)
  end else begin
   Fillchar(SockAddr,SizeOf(SockAddr),0);
   SockAddr.Family:=fmInternet;
   SockAddr.Port  :=Swap(Sender.fPort);
   SockAddr.Addr  :=Sender.ServerAddr;
   if Connect(Sender.HSocket,SockAddr,SizeOf(SockAddr))=0 then begin
    GetSocketText(Sender,wm_connect);
   end else begin
    if wsaIsBlocking then wsaCancelBlockingCall;
    ShutDown(Sender.HSocket,2);
    CloseSocket(Sender.HSocket);
    Sender.HSocket:=-1;
    PostMessage(Sender.fHandle,wm_threaddone,0,0);
    PostMessage(Sender.fHandle,wm_connect_fail,wsaGetLastError,0);
   end;
  end;
 end;

// Retreive a Message
procedure GetSocketMsg(Sender:TThrdSock; Notify:integer);
 var
  Handle :THandle;
  HSocket:THandle;
  Buffer:array[0..1024] of char;
  Size:integer;
  line:string;
  msg:TStringList;
  cr:pchar;
 begin
  Handle :=Sender.fHandle;
  HSocket:=Sender.HSocket;
  Line:='';
  repeat
   repeat Size:=recv(HSocket,Buffer,SizeOf(Buffer)-1,0) until size<>0;
   if size=-1 then begin
    PostMessage(Handle,wm_threaddone,0,0);
    PostMessage(Handle,wm_disconnect,wsaGetLastError,0);
    exit;
   end;
   Buffer[Size]:=#0;
   cr:=StrPos(Buffer,#13);
   if cr=nil then Line:=Line+Buffer;
  until cr<>nil;
  cr^:=#0; Line:=Line+Buffer;

  SendMessage(Handle,wm_setstatus,0,longint(Line));
  if (Line='')or(Line[1]='-') then begin
   PostMessage(Handle,wm_threaddone,0,0);
   PostMessage(Handle,Notify,0,0);
   exit;
  end;
  inc(cr); if cr^=#10 then inc(cr);
  StrCopy(Buffer,cr);

  msg:=TStringList.Create;
  Line:='';
  repeat
   cr:=StrPos(Buffer,#13);
   while cr<>nil do begin
    cr^:=#0;
    Line:=Line+Buffer;
    if Line='.' then begin // end of message
     SendMessage(Handle,wm_setlines,0,longint(msg));
     msg.Free;
     PostMessage(Handle,wm_threaddone,0,0);
     PostMessage(Handle,Notify,0,0);
     exit;
    end;
    Msg.Add(line);
    Line:='';
    inc(cr); if cr^=#10 then inc(cr);
    StrCopy(Buffer,cr);
    cr:=StrPos(Buffer,#13);
   end;
   Line:=Line+Buffer;
   repeat Size:=recv(HSocket,Buffer,SizeOf(Buffer)-1,0) until size<>0;
   if size=-1 then begin
    PostMessage(Handle,wm_threaddone,0,0);
    PostMessage(Handle,wm_disconnect,wsaGetLastError,0);
    exit;
   end;
   Buffer[Size]:=#0;
  until false;
 end;

function RequestMessageThread(Sender:TThrdSock):integer; stdcall;
 begin
  GetSocketMsg(Sender,wm_callback);
 end;

//-------------------------------------------
// Usefull Function
//--------------------------------------------
Function IpToLong(ip:string):LongInt;
 var
  x,i:byte;
  ipx:array[0..3] of byte;
  v:integer;
 begin
  Result:=0;
  longint(ipx):=0; i:=0;
  for x:=1 to length(ip) do
   if ip[x]='.' then begin
    inc(i);
    if i=4 then exit;
   end else begin
    if not (ip[x] in ['0'..'9']) then exit;
    v:=ipx[i]*10+ord(ip[x])-ord('0');
    if v>255 then exit;
    ipx[i]:=v;
   end;
  result:=longint(ipx);
 end;

Function CheckWinsock:boolean;
 Const
  ReadyToServe:boolean=false;
 var
  wsaData:TWsaData;
 begin
  if not ReadyToServe then ReadyToServe:=(wsaStartup($101,wsaData)=0);
  Result:=ReadyToServe;
 end;

Function HostToLong(AHost:string):LongInt;
 Type
  PHostEnt = ^THostEnt;
  THostEnt = packed record
    name     : PChar;
    aliases  : ^PChar;
    addrtype : Smallint;
    length   : Smallint;
    addr     : ^pointer;
  end;
 Var
  Host:PHostEnt;
 begin
  Host:=GetHostByName(PChar(AHost));
  if Host=nil then
   Result:=0
  else
   Result:=LongInt(Host^.Addr^^); // Wouah !
 end;

procedure split(var s1:string; sep:char; var s2:string);
 var
  p:integer;
 begin
  p:=pos(sep,s1);
  if p=0 then begin
   s2:='';
  end else begin
   s2:=copy(s1,p+1,length(s1));
   setlength(s1,p-1);
  end;
 end;

//-----------------------------------------------
// TThrdSock
//-------------------------------------------------
constructor TThrdSock.Create(AOwner:TComponent);
 begin
  inherited Create(AOwner);
  fHandle:=AllocateHWnd(WndProc);
  fLines:=TStringList.Create;
  fActive:=False;
  fEnabled:=True;
  HThread:=0;
 end;

destructor TThrdSock.Destroy;
 begin
  EOnDisconnect:=nil; // avoid Exception ???
  Active:=false;
  fLines.free;
  DeallocateHWnd(fHandle);
  inherited Destroy;
 end;

Procedure TThrdSock.SetActive(Value:boolean);
 begin
  if fActive=Value then exit;
  if Value then Connect else Disconnect;
 end;

Procedure TThrdSock.SetServer(Value:string);
 begin
  if fServer=Value then exit;
  if Active then raise Exception.Create('Unable to set Server while active');
  fServer:=Value;
 end;

Function TThrdSock.ServerAddr:longint;
 begin
  Result:=IpToLong(fServer);
  if Result=0 then begin
   Result:=HostToLong(fServer);
  end;
 end;

Procedure TThrdSock.SetPort(Value:word);
 begin
  if fPort=Value then exit;
  if Active then raise Exception.Create('Unable to set Port while active');
  fPort:=Value;
 end;

procedure TThrdSock.SetLines(Value:TStrings);
 begin
  fLines.assign(Value);
 end;

procedure TThrdSock.SetStatus(Value:string);
 begin
  fStatus:=Value;
  if Assigned(EStatusChange) then EStatusChange(Self);
 end;

Procedure TThrdSock.Connect;
 begin
  if fActive then raise Exception.Create('Allready active');
  if HThread<>0 then raise Exception.create('Unable to perform multiple connection');
  SetEnabled(False);
  CheckWinsock;
  HThread:=CreateThread(nil,0,@ConnectThread,self,0,ThreadId);
 end;

Procedure TThrdSock.Disconnect;
 begin
  if HSocket=-1 then exit;
  CancelRequest;
  ShutDown(HSocket,2);
  CloseSocket(HSocket);
  HSocket:=-1;
  fActive:=false;
  if Assigned(EOnDisconnect) then EOnDisconnect(Self);
 end;

Procedure TThrdSock.Echo(Cmd:string); // send text on active socket
 begin
  Cmd:=Cmd+#13#10;
  Send(HSocket,PChar(cmd),length(cmd),0);
 end;

Procedure TThrdSock.RequestStatus(ACallBack:TCallBack); // request a single line
 begin
  if HThread<>0 then raise Exception.Create('Socket is busy');
  SetEnabled(False);
  CallBack:=ACallBack;
  HThread:=CreateThread(nil,0,@RequestStatusThread,Self,0,ThreadId);
 end;

Procedure TThrdSock.RequestMessage(ACallBack:TCallBack); // request a message
 begin
  if HThread<>0 then raise Exception.Create('Socket is busy');
  SetEnabled(False);
  CallBack:=ACallBack;
  HThread:=CreateThread(nil,0,@RequestMessageThread,Self,0,ThreadId);
 end;

Procedure TThrdSock.CancelRequest;
 begin
  if HThread=0 then exit;
  TerminateThread(HThread,255);
  CloseHandle(HThread);
  HThread:=0;
  SetEnabled(True);
 end;

Procedure TThrdSock.SetEnabled(Value:boolean);
 var
  i:integer;
 begin
  if fEnabled=Value then exit;
  fEnabled:=Value;
  if Value then begin
   if Assigned(EOnReady) then try EOnReady(Self) except end;
   Screen.Cursor:=crDefault
  end else begin
   if Assigned(EOnBusy) then try EOnBusy(Self) except end;
   Screen.Cursor:=crAppStart;
  end;
 end;

Procedure TThrdSock.WndProc(Var Message:TMessage);
 begin
  Dispatch(Message); // dummy isn't it ? AllocateHWnd(Dispatch) doesn't work !
 end;

procedure TThrdSock.wmThreadDone(Var Msg:TMessage);
 begin
  CloseHandle(HThread);
  HThread:=0;
//  SetEnabled(True);
 end;

procedure TThrdSock.wmSetStatus(Var Msg:TMessage);
 begin
  Status:=string(Msg.lParam);
 end;

procedure TThrdSock.wmSetLines(Var Msg:TMessage);
 begin
  Lines.Assign(TStringList(Msg.lParam));
 end;

procedure TThrdSock.wmConnect(Var Msg:TMessage);
 begin
  fActive:=true;
  if Assigned(EOnconnect) then EOnConnect(Self);
  SetEnabled(True);
 end;

procedure TThrdSock.wmConnectFail(Var Msg:TMessage);
 begin
  fActive:=false;
  if Assigned(EOnconnectFail) then EOnConnectFail(Self);
 end;

procedure TThrdSock.wmDisconnect(Var Msg:TMessage);
 begin
  Active:=false;
 end;

procedure TThrdSock.wmCallBack(Var Msg:TMessage);
 var
  call:TCallBack;
 begin
  if Assigned(CallBack) then begin
   call:=CallBack;
   CallBack:=nil;
   Call;
  end;
 end;

// POP3 client -----------------------------------------------------------

Constructor TThrdPOP.Create(AOwner:TComponent);
 begin
  inherited Create(AOwner);
  Port:=110; // POP3
 end;

Procedure TThrdPOP.wmConnect(Var Msg:TMessage);
 begin
  inherited wmConnect(Msg);
  Echo('USER '+User); // Enter user-id
  RequestStatus(UserSelected); // and wait reply
 end;

Procedure TThrdPOP.UserSelected; // USER reply
 begin
//  if Assigned(EStatusChange) then EStatusChange(Self);
  if Status[1]='+' then begin // ok
   Echo('PASS '+Password); // send password
   RequestStatus(LoggedIn); // and wait reply
  end else // error ?
   SetEnabled(True);
 end;

Procedure TThrdPOP.LoggedIn; // password sent
 begin
//  if Assigned(EStatusChange) then EStatusChange(Self);
  if Status[1]='+' then begin // ok
   Echo('STAT'); // request mail status
   RequestStatus(MailStatus); // and wait reply
  end else // error ?
   SetEnabled(True);
 end;

Procedure TThrdPOP.MailStatus; // got mail status
 Var
  p:integer;
 begin
  fMailCount:=0; // default mailcount
  if Status[1]='+' then begin // ok
   p:=pos(' ',Status); if p=0 then exit;
   while (p<length(Status))and(Status[p+1] in ['0'..'9']) do begin
    inc(p);
    fMailCount:=10*fMailCount+(ord(Status[p])-ord('0'));
   end;
   if (fMailCount>0)and(Assigned(EOnMailCount)) then EOnMailCount(Self); // 0 mail ?
  end; // error ?
  SetEnabled(True);
 end;

Function TThrdPOP.GetMail(Index:integer):boolean;
 begin
  Result:=HThread=0;
  if Result then begin
   Echo('RETR '+IntToStr(Index)); // retreive message
   RequestMessage(GotMail); // and wait reply
  end;
 end;

Procedure TThrdPOP.GotMail; // got message
 begin
  SetEnabled(True);
  if Assigned(EOnMail) then EOnMail(Self);
 end;

Function TThrdPOP.DelMail(Index:integer):boolean;
 begin
  Result:=HThread=0;
  if Result then begin
   Echo('DELE '+IntToStr(Index)); // del message
   RequestStatus(MailRemoved); // wait reply
  end;
 end;

Procedure TThrdPOP.MailRemoved; // mail removed
 begin
  SetEnabled(True);
 end;

// NNTP client -----------------------------------------------------------

Constructor TThrdNNTP.Create(AOwner:TComponent);
 begin
  inherited Create(AOwner);
  Port:=119; // NNTP
  fGroups:=TStringList.Create;
  fReferences:=TStringList.Create;
  TStringList(fGroups).Sorted:=True;
 end;

Destructor TThrdNNTP.Destroy;
 begin
  fGroups.Free;
  fReferences.Free;
  inherited Destroy;
 end;

Procedure TThrdNNTP.ListGroup;
 begin
  SetEnabled(False);
  Echo('LIST');
  RequestMessage(GroupListed);
 end;

Procedure TThrdNNTP.GroupListed;
 var
  i:integer;
  p:integer;
  s:string;
 begin
  fGroups.Clear;
  for i:=0 to fLines.Count-1 do begin
   s:=fLines[i]; p:=pos(' ',s); if p>0 then SetLength(s,p-1);
   fGroups.Add(s);
  end;
  SetEnabled(True);
  if Assigned(EGotList) then EGotList(Self);
 end;

procedure TThrdNNTP.SetGroups(Value:TStrings);
 begin
  fGroups.Assign(Value);
 end;

procedure TThrdNNTP.SetGroup(Value:String);
 begin
  if fGroup=value then exit;
  fGroup:=Value;
  Echo('GROUP '+Value);
  RequestStatus(GroupStatus);
 end;

procedure TThrdNNTP.GroupStatus;
 var
  s,ss:string;
 begin
 { 201 | <count> <first> <last> <name> }
  SetEnabled(True);
  if fStatus[1]='2' then begin
   s:=fStatus;
   split(s,' ',ss); { 201 | <count> <first> <last> <name> }
   split(ss,' ',s); { <count> | <first> <last> <name> }
   fCount:=strtoint(ss);
   split(s,' ',ss); { <first> | <last> <name> }
   fFirst:=strtoint(s);
   split(ss,' ',s); { <last>  | <name> }
   fLast:=strtoint(ss);
   Status:=s;
  end else begin
   fCount:=0;
   fFirst:=0;
   fLast:=0;
  end;
  if Assigned(EOnGroupStatus) then EOnGroupStatus(Self);
 end;

Function TThrdNNTP.GetHeader(Index:integer):boolean;
 begin
  Result:=HThread=0;
  if Result then begin
   Echo('HEAD '+IntToStr(Index)); // retreive header
   RequestMessage(GotHeader); // and wait reply
  end;
 end;

Procedure TThrdNNTP.GotHeader; // got message
 var
  i:integer;
  s,ss:string;
 begin
{
Path: forums.borland.com!not-for-mail
From: DG Groep <info@dggroep.nl>
Newsgroups: borland.public.delphi.graphics
Subject: Re: Printing problem with metafiles
Date: Thu, 17 Sep 1998 13:57:52 +0200
Organization: Another Netscape Collabra Server User
Lines: 43
Message-ID: <3600F940.ADD74E21@dggroep.nl>
References: <3600D440.F7406841@dggroep.nl> <36011279.4D46@NOSPAM.ibpc.fr>
NNTP-Posting-Host: cs1p07.dial.cistron.nl
Mime-Version: 1.0
Content-Type: text/plain; charset=us-ascii
Content-Transfer-Encoding: 7bit
X-Mailer: Mozilla 4.04 [en] (WinNT; I)
}
  fReferences.Clear;
  for i:=0 to fLines.Count-1 do begin
   s:=fLines[i];
   split(s,':',ss);
   s:=uppercase(s);
   if s='FROM' then fFrom:=ss else
   if s='SUBJECT' then fSubject:=ss else
   if s='DATE' then fDate:=ss else
   if s='LINES' then fLineCount:=ss else
   if s='MESSAGE-ID' then fMessageId:=ss else
   if s='REFERENCES' then
    while ss<>'' do begin
     s:=ss;
     split(s,' ',ss);
     if s<>'' then fReferences.add(s);
    end;
  end;
  SetEnabled(True);
  if Assigned(EOnHeader) then EOnHeader(Self);
 end;

Function TThrdNNTP.GetBody:boolean;
 begin
  Result:=HThread=0;
  if Result then begin
   Echo('BODY '+fMessageID); // retreive header
   RequestMessage(GotBody); // and wait reply
  end;
 end;

Procedure TThrdNNTP.GotBody; // got message
 begin
  SetEnabled(True);
  if Assigned(EOnBody) then EOnBody(Self);
 end;

Function TThrdNNTP.GetArticle(Id:string):boolean;
 begin
  Result:=HThread=0;
  if Result then begin
   Echo('ARTICLE '+ID); // retreive article (header & body)
   RequestMessage(GotArticle); // and wait reply
  end;
 end;

Procedure TThrdNNTP.GotArticle; // got message
 var
  i:integer;
  s,ss:string;
 begin
  for i:=0 to fLines.Count-1 do begin
   s:=fLines[0];
   fLines.Delete(0);
   if s='' then break;
   split(s,':',ss);
   s:=uppercase(s);
   if s='FROM' then fFrom:=ss else
   if s='SUBJECT' then fSubject:=ss else
   if s='DATE' then fDate:=ss else
   if s='LINES' then fLineCount:=ss else
   if s='MESSAGE-ID' then fMessageId:=ss else
   if s='REFERENCES' then
    while ss<>'' do begin
     s:=ss;
     split(s,' ',ss);
     if s<>'' then fReferences.add(s);
    end;
  end;
  SetEnabled(True);
  if Assigned(EOnArticle) then EOnArticle(Self);
 end;

end.
